import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:mobile_assistant_client/constant.dart';
import 'package:flutter/material.dart';
import 'package:material_segmented_control/material_segmented_control.dart';
import 'package:mobile_assistant_client/event/back_btn_pressed.dart';
import 'package:mobile_assistant_client/event/back_btn_visibility.dart';
import 'package:mobile_assistant_client/event/delete_op.dart';
import 'package:mobile_assistant_client/event/update_delete_btn_status.dart';
import 'package:mobile_assistant_client/event/update_video_sort_order.dart';
import 'package:mobile_assistant_client/event/video_sort_menu_visibility.dart';
import 'package:mobile_assistant_client/home/video/all_video_manager_page.dart';
import 'package:mobile_assistant_client/home/video/video_folder_manager_page.dart';
import 'package:mobile_assistant_client/model/UIModule.dart';
import '../event/update_bottom_item_num.dart';
import 'package:mobile_assistant_client/util/event_bus.dart';

class VideoManagerPage extends StatefulWidget {

  @override
  State<StatefulWidget> createState() {
    return VideoManagerState();
  }
}

class VideoManagerState extends State<VideoManagerPage> {
  final _divider_line_color = Color(0xffe0e0e0);

  static final int INDEX_ALL_VIDEOS = 0;
  static final int INDEX_VIDEO_FOLDERS = 1;
  int _currentIndex = INDEX_ALL_VIDEOS;
  bool _isDeleteBtnEnabled = false;

  // 按创建时间排序
  static final int SORT_ORDER_CREATE_TIME = 1;
  // 按视频大小排序
  static final int SORT_ORDER_SIZE = 2;
  int _sortOrder = SORT_ORDER_CREATE_TIME;

  final _allVideoManagerPage = AllVideoManagerPage();
  final _videoFolderManagerPage = VideoFolderManagerPage();
  int _allItemNum = 0;
  int _selectedItemNum = 0;
  bool _isBackBtnDown = false;
  bool _isBackBtnVisible = false;
  // 排序类型菜单可见性
  bool _videoSortTypeVisible = true;

  StreamSubscription<UpdateBottomItemNum>? _updateBottomItemNumStream;
  StreamSubscription<UpdateDeleteBtnStatus>? _updateDeleteBtnStream;
  StreamSubscription<BackBtnVisibility>? _backBtnVisibilityStream;
  StreamSubscription<VideoSortMenuVisibility>? _videoSortMenuVisibilityStream;

  void _registerEventBus() {
    _updateBottomItemNumStream = eventBus.on<UpdateBottomItemNum>().listen((event) {
      updateBottomItemNumber(event.totalNum, event.selectedNum);
    });

    _updateDeleteBtnStream = eventBus.on<UpdateDeleteBtnStatus>().listen((event) {
      setDeleteBtnEnabled(event.isEnable);
    });

    _backBtnVisibilityStream = eventBus.on<BackBtnVisibility>().listen((event) {
      setState(() {
        _isBackBtnVisible = event.visible;
      });
    });
    _videoSortMenuVisibilityStream = eventBus.on<VideoSortMenuVisibility>().listen((event) {
      setState(() {
        _videoSortTypeVisible = event.visible;
      });
    });
  }

  void _unRegisterEventBus() {
    _updateDeleteBtnStream?.cancel();
    _updateBottomItemNumStream?.cancel();
    _backBtnVisibilityStream?.cancel();
    _videoSortMenuVisibilityStream?.cancel();
  }

  void setDeleteBtnEnabled(bool enable) {
    setState(() {
      _isDeleteBtnEnabled = enable;
    });
  }

  @override
  void initState() {
    super.initState();
    _registerEventBus();
  }

  @override
  Widget build(BuildContext context) {
    final pageController = PageController(initialPage: 0);

    Color getSegmentBtnColor(int index) {
      if (index == _currentIndex) {
        return Color(0xffffffff);
      } else {
        return Color(0xff5b5c62);
      }
    }

    String _getSortOrderIcon(int sortOrder) {
      if (sortOrder == SORT_ORDER_CREATE_TIME) {
        if (_sortOrder == sortOrder) {
          return "assets/icons/icon_grid_selected.png";
        } else {
          return "assets/icons/icon_grid_normal.png";
        }
      }

      if (_sortOrder == sortOrder) {
        return "assets/icons/icon_monthly_selected.png";
      } else {
        return "assets/icons/icon_monthly_normal.png";
      }
    }

    Color _getSortOrderBgColor(int sortOrder) {
      if (_sortOrder == sortOrder) {
        return Color(0xffc2c2c2);
      } else {
        return Color(0xfff5f5f5);
      }
    }

    String itemNumStr = "共${_allItemNum}项";
    if (_selectedItemNum > 0) {
      itemNumStr = "$itemNumStr (选中${_selectedItemNum}项)";
    }

    return Column(
      children: [
        Container(
          child: Stack(
            children: [
              GestureDetector(
                child: Visibility(
                  child: Align(
                    alignment: Alignment.centerLeft,
                    child: Container(
                      child: Row(
                        children: [
                          Image.asset("assets/icons/icon_right_arrow.png",
                              width: 12, height: 12),
                          Container(
                            child: Text("返回",
                                style: TextStyle(
                                    color: Color(0xff5c5c62),
                                    fontSize: 13)),
                            margin: EdgeInsets.only(left: 3),
                          ),
                        ],
                      ),
                      decoration: BoxDecoration(
                          color: _isBackBtnDown
                              ? Color(0xffe8e8e8)
                              : Color(0xfff3f3f4),
                          borderRadius:
                          BorderRadius.all(Radius.circular(3.0)),
                          border: Border.all(
                              color: Color(0xffdedede), width: 1.0)),
                      height: 25,
                      width: 50,
                      margin: EdgeInsets.only(left: 15),
                    ),
                  ),
                  visible: _isBackBtnVisible,
                ),
                onTap: () {
                  _onBackPressed();
                },
                onTapDown: (detail) {
                  setState(() {
                    _isBackBtnDown = true;
                  });
                },
                onTapCancel: () {
                  setState(() {
                    _isBackBtnDown = false;
                  });
                },
                onTapUp: (detail) {
                  setState(() {
                    _isBackBtnDown = false;
                  });
                },
              ),

              Align(
                  alignment: Alignment.center,
                  child: Container(
                    child: MaterialSegmentedControl<int>(
                      children: {
                        INDEX_ALL_VIDEOS: Container(
                          child: Text("所有视频",
                              style: TextStyle(
                                  fontSize: 12,
                                  color: getSegmentBtnColor(INDEX_ALL_VIDEOS))),
                          padding: EdgeInsets.only(left: 10, right: 10),
                        ),
                        INDEX_VIDEO_FOLDERS: Container(
                          child: Text("视频文件夹",
                              style: TextStyle(
                                  fontSize: 12,
                                  color:
                                  getSegmentBtnColor(INDEX_VIDEO_FOLDERS))),
                          padding: EdgeInsets.only(left: 10, right: 10),
                        ),
                      },
                      selectionIndex: _currentIndex,
                      borderColor: Color(0xffdedede),
                      selectedColor: Color(0xffc3c3c3),
                      unselectedColor: Color(0xfff7f5f6),
                      borderRadius: 3.0,
                      verticalOffset: 0,
                      disabledChildren: [],
                      onSegmentChosen: (index) {
                        setState(() {
                          _currentIndex = index;
                          pageController.jumpToPage(_currentIndex);
                        });
                      },
                    ),
                    height: 30,
                  )),
              Align(
                alignment: Alignment.centerRight,
                child: Container(
                  child: Row(
                    children: [
                      Visibility(
                        child: Row(
                          children: [
                            GestureDetector(
                              child: Container(
                                child: Image.asset(
                                    _getSortOrderIcon(SORT_ORDER_CREATE_TIME),
                                    width: 20,
                                    height: 20),
                                padding: EdgeInsets.fromLTRB(13, 3, 13, 3),
                                decoration: BoxDecoration(
                                  border: Border.all(
                                      color: Color(0xffdddedf), width: 1.0),
                                  borderRadius: BorderRadius.only(
                                      topLeft: Radius.circular(4.0),
                                      bottomLeft: Radius.circular(4.0)),
                                  color: _getSortOrderBgColor(SORT_ORDER_CREATE_TIME),
                                ),
                              ),
                              onTap: () {
                                if (_sortOrder != SORT_ORDER_CREATE_TIME) {
                                  setState(() {
                                    _sortOrder = SORT_ORDER_CREATE_TIME;
                                    eventBus.fire(UpdateVideoSortOrder(UpdateVideoSortOrder.TYPE_CREATE_TIME));
                                  });
                                }
                              },
                            ),
                            GestureDetector(
                              child: Container(
                                child: Image.asset(
                                    _getSortOrderIcon(SORT_ORDER_SIZE),
                                    width: 20,
                                    height: 20),
                                padding: EdgeInsets.fromLTRB(13, 3, 13, 3),
                                decoration: BoxDecoration(
                                  border: Border.all(
                                      color: Color(0xffdddedf), width: 1.0),
                                  color: _getSortOrderBgColor(SORT_ORDER_SIZE),
                                  borderRadius: BorderRadius.only(
                                      topRight: Radius.circular(4.0),
                                      bottomRight: Radius.circular(4.0)),
                                ),
                              ),
                              onTap: () {
                                if (_sortOrder != SORT_ORDER_SIZE) {
                                  setState(() {
                                    _sortOrder = SORT_ORDER_SIZE;
                                    eventBus.fire(UpdateVideoSortOrder(UpdateVideoSortOrder.TYPE_VIDEO_SIZE));
                                  });
                                }
                              },
                            ),
                          ],
                        ),
                        maintainSize: true,
                        maintainState: true,
                        maintainAnimation: true,
                        visible: _videoSortTypeVisible
                      ),
                      Container(
                          child: GestureDetector(
                            child: Opacity(
                              opacity: _isDeleteBtnEnabled ? 1.0 : 0.6,
                              child: Container(
                                child: Image.asset("assets/icons/icon_delete.png",
                                    width: 10, height: 10),
                                decoration: BoxDecoration(
                                    color: Color(0xffcb6357),
                                    border: new Border.all(
                                        color: Color(0xffb43f32), width: 1.0),
                                    borderRadius:
                                    BorderRadius.all(Radius.circular(4.0))),
                                width: 40,
                                height: 25,
                                padding:
                                EdgeInsets.fromLTRB(6.0, 4.0, 6.0, 4.0),
                              ),
                            ),
                            onTap: () {
                              debugPrint("当前删除按钮点击状态: $_isDeleteBtnEnabled");

                              if (_isDeleteBtnEnabled) {
                                eventBus.fire(DeleteOp(UIModule.Video));
                              }
                            },
                          ),
                          margin: EdgeInsets.fromLTRB(10, 0, 0, 0))
                    ],
                  ),
                  width: 160,
                ),
              )
            ],
          ),
          height: Constant.HOME_NAVI_BAR_HEIGHT,
          color: Color(0xfff6f6f6),
        ),
        Divider(color: _divider_line_color, height: 1.0, thickness: 1.0),
        Expanded(
          child: PageView(
            physics: NeverScrollableScrollPhysics(),
            children: [
              _allVideoManagerPage,
              _videoFolderManagerPage
            ],
            onPageChanged: (index) {
              debugPrint("onPageChanged, index: $index");
              setState(() {
              });
            },
            controller: pageController,
          ),
        ),
        Divider(color: _divider_line_color, height: 1.0, thickness: 1.0),
        Container(
          child: Align(
            alignment: Alignment.center,
            child: Text(itemNumStr,
                style: TextStyle(
                    fontSize: 12, color: Color(0xff646464))),
          ),
          height: 20,
          color: Color(0xfffafafa),
        )
      ],
    );
  }

  void _onBackPressed() {
    eventBus.fire(BackBtnPressed());
  }

  void updateBottomItemNumber(int allItemNum, int selectedItemNum) {
    setState(() {
      _allItemNum = allItemNum;
      _selectedItemNum = selectedItemNum;
    });
  }

  @override
  void dispose() {
    super.dispose();

    _unRegisterEventBus();
  }
}